#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 Connect to and Export from DB

 @author East Japan Accounting Center Co., Ltd

 (c)2019 The University of Aizu
 This software is released under the New BSD License.
"""
from logging import getLogger, DEBUG

import pandas as pd
import pymongo
from pymongo.errors import ConnectionFailure, InvalidName, InvalidOperation

logger = getLogger(__name__)

# Query key
KEY_DATETIME_STR = "datetime.datetime"
# Index
INDEX_DATETIME = "datetime.datetime_1"


class MongoDB:
    """
    MongoDB operation(Connection, File export, ...)
    """

    def __init__(self, host, port, db_name, collection_name=""):
        self._client = None
        self._db = None
        self._host = host
        self._port = port
        self._db_name = db_name
        self._collection_name = collection_name

    def connection(self, host="localhost", port=27017):
        """
        Connect to specified DB

        Parameters
        ----------
        host: string(Optional, Default: localhost)
            Host name on which to connect
        port: int(Optional, Default: 27017)
            Port number on which to connect

        Returns
        -------
        DB instance
        """

        try:
            self._client = pymongo.MongoClient(host=host, port=port)
        except ConnectionFailure:
            logger.error("Unable to connect DB. Please confirm the HOST and PORT.")

        return self._client

    def _export(self, db_name, collection_name, exp_path, query):
        """
        Data export

        Parameters
        ----------
        db_name: string
            DB name
        collection_name: string
            Collection name
        ext_path: string
            CSV export file path
        query: dictionary
            Query
        """

        df = self.get_document(db_name, collection_name, query)
        if df is not None:
            self._exp_csv(exp_path, df)

    def get_document(self, db_name, collection_name, query):
        """
        Data corresponding to the query

        Parameters
        ----------
        db_name: string
            DB name
        collection_name: string
            Collection name
        query: dictionary
            Query

        Returns
        -------
        df: pandas.core.frame.DataFrame
            Dataframe
        """

        self._db_name = db_name
        self._collection_name = collection_name
        df = None

        # Do NOT export if the DB, collection are not specified
        if len(db_name) == 0 and \
           len(collection_name) == 0:
            return df

        # Querying for more than one document
        try:
            self._db = self._client[db_name]
            co = self._db[collection_name]
            doc = co.find(query).sort(KEY_DATETIME_STR)
            # Convert to Dataframe
            df = pd.DataFrame.from_dict(list(doc)).astype(object)
            del df['_id']
        except InvalidName:
            logger.error("Invalid DB or collection. Please confirm.")
        except Exception:
            logger.error("Unable to get the document.")
        return df

    def read_df(self, db_name, collection_name, query=None):
        """
        Data corresponding to the query

        Parameters
        ----------
        db_name: string
            DB name
        collection_name: string
            Collection name
        query: dictionary
            Query

        Returns
        -------
        df: pandas.core.frame.DataFrame
            Dataframe
        """

        self._db_name = db_name
        self._collection_name = collection_name
        df = None

        # Do NOT export if the DB, collection are not specified
        if len(db_name) == 0 and \
           len(collection_name) == 0:
            return df

        # Querying for more than one document
        try:
            self._db = self._client[db_name]
            co = self._db[collection_name]
            if query is None:
                doc = co.find()
            else:
                doc = co.find(query)
            # Convert to Dataframe
            df = pd.DataFrame.from_dict(list(doc)).astype(object)
            del df['_id']
            del df['_class']
        except InvalidName:
            logger.error("Invalid DB or collection. Please confirm.")
        except Exception:
            logger.error("Unable to get the document.")
        return df

    def read_dict(self, db_name, collection_name, query=None, latest=None, limit=0):
        """
        Data corresponding to the query

        Parameters
        ----------
        db_name: string
            DB name
        collection_name: string
            Collection name
        query: dictionary
            Query

        Returns
        -------
        df: pandas.core.frame.DataFrame
            Dataframe
        """

        self._db_name = db_name
        self._collection_name = collection_name
        doc = None
        set_sort = pymongo.ASCENDING
        # Do NOT export if the DB, collection are not specified
        if len(db_name) == 0 and \
           len(collection_name) == 0:
            return doc

        if latest is not None:
            if latest:
                set_sort = pymongo.DESCENDING
            else:
                set_sort = pymongo.ASCENDING

        # Querying for more than one document
        try:
            self._db = self._client[db_name]
            co = self._db[collection_name]
            if query is None:
                doc = co.find().sort("_id", set_sort)
            else:
                doc = co.find(query).sort("_id", set_sort)
            if 0 < limit:
                doc = doc.limit(limit)
        except InvalidName:
            logger.error("Invalid DB or collection. Please confirm.")
        except Exception:
            logger.error("Unable to get the document.")
        return list(doc)

    def insert_document(self, db_name, collection_name, doc):
        """
        Insert the one document in the specified DB

        Parameters
        ----------
        db_name: string
            DB name
        collection_name: string
            Collection name
        doc: dictionary
            Document

        Returns
        -------
        True: Success False: Failure
        """

        self._db_name = db_name
        self._collection_name = collection_name
        rtn = False

        # Do NOT export if the DB, collection are not specified
        if len(db_name) == 0 or len(collection_name) == 0:
            logger.error("DB param is empty.")
            return rtn

        # Insert the one document in the specified DB
        try:
            self._db = self._client[db_name]
            co = self._db[collection_name]
            ins_one_result = co.insert_one(doc)
            if ins_one_result.acknowledged:
                rtn = True
        except InvalidOperation:
            logger.error("Invalid DB operation. Please confirm.")
        except Exception as e:
            logger.error(e)
            logger.error("Unable to insert the document.")

        return rtn

    def _exp_csv(self, exp_path, df):
        """
        Export as a CSV file

        Parameters
        ----------
        ext_path: string
            CSV export file path
        df: pandas.core.frame.DataFrame
            Dataframe

        Returns
        -------
        rtn: boolean
            Export result
            True: Success
            False: Failure
        """

        rtn = False

        if len(exp_path) == 0:
            logger.error("Export filepath is empty.")
            return rtn

        try:
            df.to_csv(exp_path, index=False)
            rtn = True
        except Exception:
            logger.error("Export file create error")

        return rtn

    def _create_index(self):
        """
        Create the DB index
        """

        if self._client:
            co = self._db[self._collection_name]
            index_list = co.index_information()
            if INDEX_DATETIME not in index_list:
                co.create_index([(KEY_DATETIME_STR, pymongo.ASCENDING)])
                logger.info("Create the index.")

    def close(self):
        """
        MongoDB client close
        """

        if self._client:
            self._client.close()
