#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- Python -*-

"""
 rdr_sub_main.py

 @author East Japan Accounting Center Co., Ltd

 (c)2021 The University of Aizu
 This software is released under the New BSD License.
"""
import copy
import time
from datetime import datetime, timezone, timedelta
from logging import getLogger, basicConfig
from logging import DEBUG, INFO
import os
import mongodb
import queue
import threading
import json
import pandas as pd
import re
from rdr_mqtt import RdrMqtt


JST = timezone(timedelta(hours=+9), 'JST')
OBJECT_ABSOLUTE_ROOT_DIR = "/home/robot-db/var/rdr"
OBJECT_RELATIVE_ROOT_DIR = "db"

# File R/W
FILE_MODE_WRITE = "w"
EXT_LOG = ".log"
APP_LOG_ROOT_DIR = "log/"

# Setup logger
log_fmt = "%(asctime)s %(name)s - %(levelname)s - %(message)s :fn[%(module)s:%(funcName)s]"
curr_datetime = datetime.now()
curr_date = "{0:04d}{1:02d}{2:02d}".format(curr_datetime.year, curr_datetime.month, curr_datetime.day)
curr_time = "{0:02d}{1:02d}{2:02d}".format(curr_datetime.hour, curr_datetime.minute, curr_datetime.second)
log_root_dir = (APP_LOG_ROOT_DIR + curr_date + "/")
if os.path.isdir(log_root_dir) is False:
    os.makedirs(log_root_dir)
log_filename = "{0}{1}_{2}_{3}{4}".format(log_root_dir, os.path.splitext(os.path.basename(__file__))[0], curr_date, curr_time, EXT_LOG)
basicConfig(filename=log_filename, filemode=FILE_MODE_WRITE, level=DEBUG, format=log_fmt)
logger = getLogger(__name__)

TOPIC_HIERARCHY_BOUNDARY = 2
UPDATE_TIME = 5


class LifeTime:
    thread_life_time = False

    @classmethod
    def is_life_time(cls):
        return cls.thread_life_time

    @classmethod
    def set_life_time_activation(cls):
        cls.thread_life_time = True

    @classmethod
    def set_life_time_deactivation(cls):
        cls.thread_life_time = False


def cast_value(value, data_type):
    if data_type == 0:  # 文字列
        return str(value)
    elif data_type == 1:  # 数字
        return int(value)
    else:
        raise ValueError("Unsupported data type")


def create_query(df: pd.DataFrame):
    """
    MongoDB用のクエリを構築する関数。

    Parameters:
        df (pd.DataFrame): 検索条件が格納されたDataFrame。

    Returns:
        dict: MongoDB用のクエリ。
    """
    filters = df.get("filter")
    relatedness = df.get("related")

    # 条件に対応する比較演算子
    condition_mapping = {
        0: "$eq",   # =
        1: "$gte",  # >=
        2: "$gt",   # >
        3: "$lte",  # <=
        4: "$lt",   # <
        5: "$ne"    # !=
    }

    # データ形式変換

    # フィルタ条件を作成
    filter_queries = {}
    for f in filters:
        field_name = f["name"]
        if len(field_name) == 0:
            continue
        operator = condition_mapping[f["condition"]]
        value = cast_value(f["value"], f["type"])
        filter_queries[field_name] = {operator: value}

    # related の条件を考慮してクエリを統合
    query = []
    for related in relatedness:
        condition_1 = filters[related["subject_id_1"]]
        condition_2 = filters[related["subject_id_2"]]
        relation_operator = "$or" if related["operator"] == 1 else "$and"

        sub_query = {
            relation_operator: [
                {condition_1["name"]: {condition_mapping[condition_1["condition"]]: cast_value(condition_1["value"],
                                                                                               condition_1["type"])}},
                {condition_2["name"]: {
                    condition_mapping[condition_2["condition"]]: cast_value(condition_2["value"], condition_2["type"])}}
            ]
        }
        query.append(sub_query)

    # 最終クエリを構築
    if len(query) > 1:
        final_query = {"$or": query}
    else:
        if query:
            final_query = query[0]
        elif any(filter_queries):
            final_query = filter_queries
        else:
            final_query = None
    logger.info(final_query)
    return final_query


def publish_main_thread(db_read_queue: queue.Queue, publish_topic_df: pd.DataFrame, subscribe_topic_df: pd.DataFrame,db_param: dict, error_queue: queue.Queue):
    cycle_a = set(publish_topic_df["cycle"].tolist())
    count_cycle_dict = {key: 0 for key in cycle_a}
    df_update_time = datetime.now() + timedelta(minutes=UPDATE_TIME)
    try:
        while LifeTime.is_life_time():
            count_cycle_dict = {key: count_cycle_dict.get(key)+1 for key in count_cycle_dict.keys()}
            publish_count_a = [v for k, v in count_cycle_dict.items() if k == v]
            publish_df = publish_topic_df[publish_topic_df["cycle"].isin(publish_count_a)]
            for index, row in publish_df.iterrows():
                row_dict = row.to_dict()
                topic_name = row_dict.get("name")
                hit_df = subscribe_topic_df[subscribe_topic_df["topic_id"] == row_dict.get("reference_id")]
                hit_df = hit_df.reset_index(drop=True)
                collection_name = hit_df.at[0, "name"].replace('/', '_')
                limit = row_dict.get("limit")
                latest = row_dict.get("latest")
                query = create_query(row_dict)
                set_queue = [topic_name, collection_name, query, latest, limit]
                db_read_queue.put(set_queue, block=True, timeout=None)
            for k, v in count_cycle_dict.items():
                if k == v:
                    count_cycle_dict[k] = 0
            time.sleep(1)
            if df_update_time < datetime.now():
                subscribe_topic_df, publish_topic_df = read_topic_list(db_param)
                new_cycle_a = set(publish_topic_df["cycle"].tolist())
                diff_cycle_a = list(set(new_cycle_a) - set(cycle_a))
                for diff_cycle in diff_cycle_a:
                    count_cycle_dict[diff_cycle] = 0

    except Exception as e:
        error_queue.put(True)
        logger.error(e)


def db_read_thread(db_queue: queue.Queue, publish_queue: queue.Queue, db_param: dict, error_queue: queue.Queue):
    # Connect to DB
    mongo = mongodb.MongoDB(db_param['host'], db_param['port'], db_param['public_db_name'])
    db_name = db_param['public_db_name']
    try:
        with mongo.connection():
            while LifeTime.is_life_time():
                if not db_queue.empty():
                    db_req = db_queue.get(block=True, timeout=None)
                    topic_name = db_req[0]
                    collection_name = db_req[1]
                    query = db_req[2]
                    latest = db_req[3]
                    limit = db_req[4]
                    logger.info("DB received collection:[{}]".format(collection_name))
                    dict_a = mongo.read_dict(db_name, collection_name, query, latest, limit)
                    publish_message = []
                    for elem in dict_a:
                        elem.pop("_id")
                        publish_message.append(elem)
                    set_queue = [topic_name, publish_message]
                    db_queue.task_done()
                    publish_queue.put(set_queue, block=True, timeout=None)
    except Exception as e:
        error_queue.put(True)
        logger.error(e)


def db_write_thread(db_queue: queue.Queue, db_param: dict, error_queue: queue.Queue):
    # Connect to DB
    mongo = mongodb.MongoDB(db_param['host'], db_param['port'], db_param['public_db_name'])
    try:
        with mongo.connection():
            while LifeTime.is_life_time():
                if not db_queue.empty():
                    db_req = db_queue.get(block=True, timeout=None)
                    db_name = db_param['public_db_name']
                    collection_name = db_req[0]
                    doc = db_req[1]
                    logger.info("DB received collection:[{}]".format(collection_name))
                    mongo.insert_document(db_name, collection_name, doc)
                    db_queue.task_done()
    except Exception as e:
        error_queue.put(True)
        logger.error(e)


class MqttSub(RdrMqtt):
    def __init__(self, mqtt_queue: queue.Queue):
        super(MqttSub, self).__init__()
        self.message_queue = mqtt_queue

    def on_message(self, mqttc, obj, msg):
        """
        Called when a message has been received on a topic that the client subscribes to.
        See detail https://github.com/eclipse/paho.mqtt.python#callbacks

        Parameters
        ----------
        mqttc
            the client instance for this callback
        obj
            the private user data as set in Client() or user_data_set()
        msg
            an instance of MQTTMessage

        Returns
        -------
        None
        """
        try:
            dict_message = json.loads(msg.payload.decode("utf-8"))
            self._lock.acquire()
            logger.info("MQTT received topic:[{}]".format(msg.topic))
            put_data = [msg.topic, dict_message]
            self.message_queue.put(put_data, block=True, timeout=None)
            self._lock.release()
        except json.JSONDecodeError:
            logger.error("mqtt message decode error")


def mqtt_publish_thread(mqtt_queue: queue.Queue, error_queue: queue.Queue):
    mqtt_ins = RdrMqtt()
    if mqtt_ins.public_connection():
        try:
            while LifeTime.is_life_time():
                if not mqtt_queue.empty():
                    mqtt_msg_a = mqtt_queue.get(block=True, timeout=None)
                    receive_topic_name = mqtt_msg_a[0]
                    receive_msg = mqtt_msg_a[1]
                    mqtt_ins.publish_send(receive_msg, receive_topic_name)
                    mqtt_queue.task_done()
        except Exception as e:
            error_queue.put(True)
            logger.error(e)
        finally:
            mqtt_ins.disconnect()
    else:
        error_queue.put(True)


def mqtt_subscribe_thread(db_queue: queue.Queue, mqtt_queue: queue.Queue, topics, db_param, error_queue: queue.Queue):
    mqtt_ins = MqttSub(mqtt_queue)
    if mqtt_ins.subscribe_register(topics=topics):
        registered_topics = copy.copy(topics)
        try:
            df_update_time = datetime.now() + timedelta(minutes=UPDATE_TIME)
            while LifeTime.is_life_time():
                if not mqtt_queue.empty():
                    mqtt_msg_a = mqtt_queue.get(block=True, timeout=None)
                    # Extract the collection name
                    receive_topic_name = mqtt_msg_a[0]
                    receive_msg = mqtt_msg_a[1]
                    # convert message
                    collection_name = receive_topic_name.replace('/', '_')
                    if receive_msg is not None:
                        set_db_msg = [collection_name, receive_msg]
                        logger.info("set DB message collection:[{}]".format(collection_name))
                        db_queue.put(set_db_msg, block=True, timeout=None)
                    mqtt_queue.task_done()
                if df_update_time < datetime.now():
                    subscribe_topic_df, _ = read_topic_list(db_param)
                    new_topic_list = list(set(subscribe_topic_df["name"].tolist()) - set(registered_topics))
                    old_topic_list = list(set(registered_topics) - set(subscribe_topic_df["name"].tolist()))
                    if 0 < len(new_topic_list):
                        set_topics = [(set_topic, 0) for set_topic in new_topic_list]
                        mqtt_ins.subscribe(set_topics)
                        logger.info("add subscribe: {}".format(new_topic_list))
                    if 0 < len(old_topic_list):
                        mqtt_ins.unsubscribe(old_topic_list)
                        logger.info("unsubscribe: {}".format(old_topic_list))
                    registered_topics = subscribe_topic_df["name"].tolist()
                    df_update_time = datetime.now() + timedelta(minutes=1)
        except Exception as e:
            error_queue.put(True)
            logger.error(e)
        finally:
            mqtt_ins.disconnect()
    else:
        error_queue.put(True)


def read_topic_list(db_param):
    mongo = mongodb.MongoDB(db_param['host'], db_param['port'], db_param['db_name'])
    try:
        with mongo.connection():
            topic_df = mongo.read_df(db_param['db_name'], db_param['topic_collection'])
        if topic_df is None:
            subscribe_topic_df = None
            publish_topic_df = None
            return subscribe_topic_df, publish_topic_df
        subscribe_topic_df = topic_df[topic_df["type"]==False]
        publish_topic_df = topic_df[topic_df["type"]==True]
    except OSError:
        subscribe_topic_df = None
        publish_topic_df = None
    except Exception:
        subscribe_topic_df = None
        publish_topic_df = None
    return subscribe_topic_df, publish_topic_df


def main():

    db_write_queue = queue.Queue()
    db_read_queue = queue.Queue()
    mqtt_subscribe_queue = queue.Queue()
    mqtt_publish_queue = queue.Queue()
    error_queue = queue.Queue()
    result, db_param = RdrMqtt.is_json_file_parse("./config/db_settings.json")
    if result is False:
        return
    subscribe_topic_df, publish_topic_df = read_topic_list(db_param)
    subscribe_topic_list = subscribe_topic_df["name"].tolist()
    db_write_thread_ins = threading.Thread(target=db_write_thread, args=(db_write_queue, db_param, error_queue), daemon=True)
    mqtt_subscribe_thread_ins = threading.Thread(target=mqtt_subscribe_thread, args=(db_write_queue, mqtt_subscribe_queue, subscribe_topic_list, db_param, error_queue), daemon=True)

    publish_main_thread_ins = threading.Thread(target=publish_main_thread, args=(db_read_queue, publish_topic_df, subscribe_topic_df, db_param, error_queue), daemon=True)
    db_read_thread_ins = threading.Thread(target=db_read_thread, args=(db_read_queue, mqtt_publish_queue, db_param, error_queue), daemon=True)
    mqtt_publish_thread_ins = threading.Thread(target=mqtt_publish_thread, args=(mqtt_publish_queue, error_queue), daemon=True)

    try:
        LifeTime.set_life_time_activation()
        mqtt_subscribe_thread_ins.start()
        db_write_thread_ins.start()
        mqtt_publish_thread_ins.start()
        db_read_thread_ins.start()
        publish_main_thread_ins.start()

        while True:
            if not error_queue.empty():
                LifeTime.set_life_time_deactivation()
                break
    except KeyboardInterrupt:
        LifeTime.set_life_time_deactivation()
        time.sleep(5)
    except Exception as e:
        logger.error(e)

if __name__ == "__main__":
    main()
